from pathlib import Path

import dash
import dash_bootstrap_components as dbc
import diskcache
import pandas as pd
from dash import ALL, MATCH, Input, Output, State, dcc, html
from dash.exceptions import PreventUpdate
from dash.long_callback import DiskcacheLongCallbackManager

from metaDMG.viz import content, figures, viz_utils
from metaDMG.viz.results import VizResults


def get_app(results_dir):

    ## Diskcache
    cache = diskcache.Cache("./.cache")
    long_callback_manager = DiskcacheLongCallbackManager(cache)

    viz_results = VizResults(results_dir)

    #%%

    viz_utils.set_custom_theme()

    #%%

    external_stylesheets = [dbc.themes.BOOTSTRAP]
    # external_stylesheets = [dbc.themes.ZEPHYR]
    # external_scripts = [
    #     "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.4/"
    #     "MathJax.js?config=TeX-MML-AM_CHTML",
    # ]

    app = dash.Dash(
        __name__,
        external_stylesheets=external_stylesheets,
        # external_scripts=external_scripts, # Not needed with Dash 3.2, wuhuu!
        title="metaDMG",
        update_title="Updating...",
    )

    # to allow custom css
    # app.scripts.config.serve_locally = True

    #%%

    # (1) No sidebars, (2) Only left filter sidebar,
    # (3) Only right plot sidebar, (4) Both sidebars
    start_configuration_id = 3

    configurations = viz_utils.get_configurations(
        sidebar_left_width=30,
        sidebar_right_width=30,
        content_main_margin=1,
    )
    start_configuration = configurations[start_configuration_id]

    d_columns_latex, columns, columns_no_log = viz_utils.get_d_columns_latex(
        viz_results
    )

    #%%

    app.layout = content.get_app_layout(viz_results, start_configuration)

    #%%

    @app.callback(
        output=[
            Output(
                "navbar_collapsed_toggle_styling",
                "is_open",
            ),
            Output(
                "navbar_btn_toggle_styling",
                "outline",
            ),
        ],
        inputs=dict(
            n=Input(
                "navbar_btn_toggle_styling",
                "n_clicks",
            ),
            is_open=State(
                "navbar_collapsed_toggle_styling",
                "is_open",
            ),
        ),
    )
    def toggle_styling(n, is_open):
        # after click
        if n:
            return not is_open, is_open
        # initial setup
        return is_open, True

    #%%

    @app.callback(
        output=[
            Output(
                "sidebar_right_collapsed_toggle_combined",
                "is_open",
            ),
            Output(
                "sidebar_right_btn_toggle_combined",
                "outline",
            ),
        ],
        inputs=dict(
            n=Input(
                "sidebar_right_btn_toggle_combined",
                "n_clicks",
            ),
            is_open=State(
                "sidebar_right_collapsed_toggle_combined",
                "is_open",
            ),
        ),
    )
    def toggle_sidebar_right_combined(n, is_open):
        if n:
            return not is_open, is_open
        return is_open, False

    @app.callback(
        output=[
            Output(
                "sidebar_right_collapsed_toggle_results",
                "is_open",
            ),
            Output(
                "sidebar_right_btn_toggle_results",
                "outline",
            ),
        ],
        inputs=dict(
            n=Input(
                "sidebar_right_btn_toggle_results",
                "n_clicks",
            ),
            is_open=State(
                "sidebar_right_collapsed_toggle_results",
                "is_open",
            ),
        ),
    )
    def toggle_sidebar_right_results(n, is_open):
        if n:
            return not is_open, is_open
        return is_open, True

    @app.callback(
        output=[
            Output(
                "sidebar_right_collapsed_toggle_forward_reverse",
                "is_open",
            ),
            Output(
                "sidebar_right_btn_toggle_forward_reverse",
                "outline",
            ),
        ],
        inputs=dict(
            n=Input(
                "sidebar_right_btn_toggle_forward_reverse",
                "n_clicks",
            ),
            is_open=State(
                "sidebar_right_collapsed_toggle_forward_reverse",
                "is_open",
            ),
        ),
    )
    def toggle_sidebar_right_forward_reverse(n, is_open):
        if n:
            return not is_open, is_open
        return is_open, True

    #%%

    @app.callback(
        output=Output(
            "sidebar_right_graph_combined",
            "figure",
        ),
        inputs=Input(
            "main_graph",
            "clickData",
        ),
    )
    def update_sidebar_right_plot_combined(click_data):
        return figures.update_raw_count_plots(
            viz_results,
            click_data,
            forward_reverse="",
        )

    @app.callback(
        output=Output(
            "sidebar_right_graph_forward",
            "figure",
        ),
        inputs=Input(
            "main_graph",
            "clickData",
        ),
    )
    def update_sidebar_right_plot_forward(click_data):
        return figures.update_raw_count_plots(
            viz_results,
            click_data,
            forward_reverse="Forward",
        )

    @app.callback(
        output=Output(
            "sidebar_right_graph_reverse",
            "figure",
        ),
        inputs=Input(
            "main_graph",
            "clickData",
        ),
    )
    def update_sidebar_right_plot_reverse(click_data):
        return figures.update_raw_count_plots(
            viz_results,
            click_data,
            forward_reverse="Reverse",
        )

    @app.callback(
        output=Output(
            "sidebar_right_datatable_results",
            "children",
        ),
        inputs=Input(
            "main_graph",
            "clickData",
        ),
    )
    def update_sidebar_right_datatable_results(click_data):
        if click_data:
            sample, tax_id = viz_utils.get_sample_tax_id_from_click_data(
                viz_results, click_data
            )

            df_fit = viz_results.filter({"sample": sample, "tax_id": tax_id})
            if len(df_fit) != 1:
                raise AssertionError(f"Should only be length 1")

            ds = df_fit.iloc[0]

            forward_only = []
            if viz_results.forward_only:
                forward_only = ["Forward only!", html.Br(), html.Br()]

            bayesian_list = []
            if viz_results.Bayesian:

                s1 = "Bayesian_D_max_confidence_interval_1_sigma_low"
                s2 = "Bayesian_D_max_confidence_interval_1_sigma_high"
                if s1 in ds and s2 in ds:

                    s_D_max = f", [{ds[s1]:.3f}, {ds[s2]:.3f}]"

                    conf_q_low = ds["Bayesian_q_confidence_interval_1_sigma_low"]
                    conf_q_high = ds["Bayesian_q_confidence_interval_1_sigma_high"]
                    s_q = f", [{conf_q_low:.3f}, {conf_q_high:.3f}]"

                    conf_phi_low = viz_utils.human_format(
                        ds["Bayesian_phi_confidence_interval_1_sigma_low"]
                    )
                    conf_phi_high = viz_utils.human_format(
                        ds["Bayesian_phi_confidence_interval_1_sigma_high"]
                    )
                    s_phi = f", [{conf_phi_low}, {conf_phi_high}]"

                    conf_A_low = ds["Bayesian_A_confidence_interval_1_sigma_low"]
                    conf_A_high = ds["Bayesian_A_confidence_interval_1_sigma_high"]
                    s_A = f", [{conf_A_low:.3f}, {conf_A_high:.3f}]"

                    conf_c_low = ds["Bayesian_c_confidence_interval_1_sigma_low"]
                    conf_c_high = ds["Bayesian_c_confidence_interval_1_sigma_high"]
                    s_c = f", [{conf_c_low:.3f}, {conf_c_high:.3f}]"

                else:
                    s_D_max = ""
                    s_q = ""
                    s_phi = ""
                    s_c = ""

                bayesian_list = [
                    "Fit results:",
                    html.Br(),
                    f"D-max: {ds['Bayesian_D_max']:.3f} "
                    f"± {ds['Bayesian_D_max_std']:.3f}" + s_D_max,
                    html.Br(),
                    f"significance: {ds['Bayesian_significance']:.2f}",
                    html.Br(),
                    f"P(>1%): {ds['Bayesian_prob_gt_1p_damage']:.2%}",
                    html.Br(),
                    f"P(>0%): {ds['Bayesian_prob_not_zero_damage']:.2%}",
                    html.Br(),
                    f"q: {ds['Bayesian_q']:.3f} " f"± {ds['Bayesian_q_std']:.3f}" + s_q,
                    html.Br(),
                    f"phi: {viz_utils.human_format(ds['Bayesian_phi'])} "
                    f"± {viz_utils.human_format(ds['Bayesian_phi_std'])}" + s_phi,
                    html.Br(),
                    f"A: {ds['Bayesian_A']:.3f} " f"± {ds['Bayesian_A_std']:.3f}" + s_A,
                    html.Br(),
                    f"c: {ds['Bayesian_c']:.3f} " f"± {ds['Bayesian_c_std']:.3f}" + s_c,
                    html.Br(),
                    f"rho Ac: {ds['Bayesian_rho_Ac']:.3f}",
                    html.Br(),
                    html.Br(),
                ]

            # fmt: off
            lines = [
                html.Br(),
                f"Sample: {ds['sample']}", html.Br(), html.Br(),

                f"Tax Name: {ds['tax_name']}",
                html.Br(),
                f"Tax rank: {ds['tax_rank']}",
                html.Br(),
                f"Tax ID: {ds['tax_id']}",
                html.Br(),
                html.Br(),

                f"N reads: {viz_utils.human_format(ds['N_reads'])}",
                html.Br(),
                f"N alignments: {viz_utils.human_format(ds['N_alignments'])}",
                html.Br(),
                f"Readlength: {ds['mean_L']:.1f} ± {ds['std_L']:.1f}",
                html.Br(),
                f"GC: {ds['mean_GC']:.2f} ± {ds['std_GC']:.2f}",
                html.Br(),
                html.Br(),

                *forward_only,

                *bayesian_list,

                "MAP results:",
                html.Br(),
                f"D-max: {ds['D_max']:.3f} "
                f"± {ds['D_max_std']:.3f}",
                html.Br(),
                f"significance: {ds['significance']:.2f}",
                html.Br(),
                f"q: {ds['q']:.3f} ± {ds['q_std']:.3f}",
                html.Br(),
                f"phi: {viz_utils.human_format(ds['phi'])} "
                f"± {viz_utils.human_format(ds['phi_std'])}",
                html.Br(),
                f"A: {ds['A']:.3f} ± {ds['A_std']:.3f}",
                html.Br(),
                f"c: {ds['c']:.3f} ± {ds['c_std']:.3f}",
                html.Br(),
                f"asymmetry: {ds['asymmetry']:.3f}",
                html.Br(),
                f"rho Ac: {ds['rho_Ac']:.3f}",
                html.Br(),
                f"Fit valid: {ds['valid']}",
                html.Br(),
                html.Br(),

                f"Tax path: {viz_utils.tax_path_to_string(ds['tax_path'])}",
                html.Br(),
                html.Br(),

                # f"LR_All: {ds['LR_All']:.3f}", html.Br(),
                # f"LR_ForRev: {ds['LR_ForRev']:.3f}", html.Br(),
                # f"LR_ForRev_All: {ds['LR_ForRev_All']:.3f}", html.Br(), html.Br(),
                # f"chi2 all: {ds['chi2_all']:.3f}", html.Br(),
                # f"chi2_ForRev: {ds['chi2_ForRev']:.3f}", html.Br(),
            ]
            # fmt: on

            return [
                html.P(lines),
                # html.Div(
                #     children=html.Img(
                #         src="http://latex2png.com/pngs/2cc065b79dc0045babf538323a9a59c4.png",
                #         style={
                #             "maxWidth": "100%",
                #             "maxHeight": "100%",
                #             "marginLeft": "auto",
                #             "marginRight": "auto",
                #         },
                #     ),
                # ),
            ]

        return html.P(["Please select a point"])

    #%%

    @app.callback(
        output=[
            Output(
                "main_graph",
                "figure",
            ),
            Output(
                "modal",
                "is_open",
            ),
        ],
        inputs=dict(
            sidebar_left_dropdown_samples=Input(
                "sidebar_left_dropdown_samples",
                "value",
            ),
            sidebar_left_taxa_input_specific=Input(
                "sidebar_left_taxa_input_specific",
                "value",
            ),
            sidebar_left_taxa_input_path_contains=Input(
                "sidebar_left_taxa_input_path_contains",
                "value",
            ),
            # tax_id_plot_button=Input("tax_id_plot_button", "n_clicks",),
            sidebar_left_results_dynamic_value=Input(
                {"type": "sidebar_left_results_dynamic", "index": ALL},
                "value",
            ),
            xaxis_column_name=Input(
                "xaxis_column",
                "value",
            ),
            yaxis_column_name=Input(
                "yaxis_column",
                "value",
            ),
            marker_transformation_variable=Input(
                "marker_transformation_variable",
                "value",
            ),
            marker_transformation_function=Input(
                "marker_transformation_function",
                "value",
            ),
            marker_transformation_slider=Input(
                "marker_transformation_slider",
                "value",
            ),
            modal_close_button=Input(
                "modal_close_button",
                "n_clicks",
            ),
            sidebar_left_results_dynamic_ids=State(
                {"type": "sidebar_left_results_dynamic", "index": ALL},
                "id",
            ),
            modal=State(
                "modal",
                "is_open",
            ),
        ),
    )
    def update_main_graph(
        sidebar_left_dropdown_samples,
        sidebar_left_taxa_input_specific,
        sidebar_left_taxa_input_path_contains,
        sidebar_left_results_dynamic_value,
        xaxis_column_name,
        yaxis_column_name,
        marker_transformation_variable,
        marker_transformation_function,
        marker_transformation_slider,
        modal_close_button,
        sidebar_left_results_dynamic_ids,
        modal,
    ):

        # if modal is open and the "close" button is clicked, close down modal
        if modal_close_button and modal:
            return dash.no_update, False

        # if no files selected
        if not sidebar_left_dropdown_samples:
            raise PreventUpdate

        viz_results.set_marker_size(
            marker_transformation_variable,
            marker_transformation_function,
            marker_transformation_slider,
        )

        df_results_filtered = filter_dataframe(
            viz_results,
            sidebar_left_dropdown_samples,
            sidebar_left_results_dynamic_ids,
            sidebar_left_results_dynamic_value,
            sidebar_left_taxa_input_specific,
            sidebar_left_taxa_input_path_contains,
            # sidebar_left_tax_id_subspecies,
        )

        # raise modal warning if no results due to too restrictive filtering
        if len(df_results_filtered) == 0:
            return dash.no_update, True

        fig = figures.make_figure(
            viz_results,
            df=df_results_filtered,
            xaxis_column_name=xaxis_column_name,
            yaxis_column_name=yaxis_column_name,
            d_columns_latex=d_columns_latex,
        )

        return fig, dash.no_update

    #%%

    @app.callback(
        output=Output(
            "sidebar_left_dropdown_samples",
            "value",
        ),
        inputs=Input(
            "sidebar_left_dropdown_samples",
            "value",
        ),
    )
    def update_dropdown_samples_when_Select_all(sidebar_left_dropdown_samples):
        if viz_utils.key_is_in_list_case_insensitive(
            sidebar_left_dropdown_samples,
            "Select all",
        ):
            sidebar_left_dropdown_samples = viz_results.samples
        elif viz_utils.key_is_in_list_case_insensitive(
            sidebar_left_dropdown_samples,
            "Default selection",
        ):
            sidebar_left_dropdown_samples = viz_utils.get_samples_each(
                viz_results.samples
            )

        sidebar_left_dropdown_samples = list(sorted(sidebar_left_dropdown_samples))

        return sidebar_left_dropdown_samples

    #%%

    @app.callback(
        output=Output(
            "sidebar_left_results_container",
            "children",
        ),
        inputs=dict(
            dropdown_names=Input(
                "sidebar_left_results",
                "value",
            ),
            children=State(
                "sidebar_left_results_container",
                "children",
            ),
            current_ids=State(
                {"type": "sidebar_left_results_dynamic", "index": ALL},
                "id",
            ),
        ),
        prevent_initial_call=True,
    )
    def update_sidebar_left_fit_result_sliders(dropdown_names, children, current_ids):

        id_type = "dbc"

        current_names = content.get_current_names(current_ids)

        # add new slider
        if content.slider_is_added(current_names, dropdown_names):
            column = content.get_name_of_added_slider(current_names, dropdown_names)
            new_element = content.make_new_slider(viz_results, column, id_type=id_type)
            children.append(new_element)

        # remove selected slider
        else:
            columns = content.get_name_of_removed_slider(current_names, dropdown_names)
            for column in columns:
                content.remove_name_from_children(column, children, id_type=id_type)

        return children

    @app.callback(
        output=Output(
            {"type": "sidebar_left_results_dynamic_name", "index": MATCH},
            "children",
        ),
        inputs=dict(
            dynamic_slider_values=Input(
                {"type": "sidebar_left_results_dynamic", "index": MATCH},
                "value",
            ),
            sidebar_left_results_dynamic_name=State(
                {"type": "sidebar_left_results_dynamic", "index": MATCH},
                "id",
            ),
        ),
        prevent_initial_call=False,
    )
    def update_sidebar_left_fit_result_slider_names(
        dynamic_slider_values, sidebar_left_results_dynamic_name
    ):
        column = sidebar_left_results_dynamic_name["index"]
        name = content.get_slider_name(column, dynamic_slider_values)
        return name

    #%%

    @app.callback(
        output=[
            Output(
                "sidebar_left_samples_collapsed",
                "is_open",
            ),
            Output(
                "sidebar_left_samples_btn",
                "outline",
            ),
        ],
        inputs=dict(
            n=Input(
                "sidebar_left_samples_btn",
                "n_clicks",
            ),
            is_open=State(
                "sidebar_left_samples_collapsed",
                "is_open",
            ),
        ),
    )
    def toggle_sidebar_left_samples(n, is_open):
        # after click
        if n:
            return not is_open, is_open
        # initial setup
        return True, False

    @app.callback(
        output=[
            Output(
                "sidebar_left_taxanomics_collapsed",
                "is_open",
            ),
            Output(
                "sidebar_left_taxanomics_btn",
                "outline",
            ),
        ],
        inputs=dict(
            n=Input(
                "sidebar_left_taxanomics_btn",
                "n_clicks",
            ),
            is_open=State(
                "sidebar_left_taxanomics_collapsed",
                "is_open",
            ),
        ),
    )
    def toggle_sidebar_left_taxanomics(n, is_open):
        if n:
            return not is_open, is_open
        return False, True

    @app.callback(
        output=[
            Output(
                "sidebar_left_results_collapsed",
                "is_open",
            ),
            Output(
                "sidebar_left_results_btn",
                "outline",
            ),
        ],
        inputs=dict(
            n=Input(
                "sidebar_left_results_btn",
                "n_clicks",
            ),
            is_open=State(
                "sidebar_left_results_collapsed",
                "is_open",
            ),
        ),
    )
    def toggle_sidebar_left_results(n, is_open):
        if n:
            return not is_open, is_open
        return True, False

    #%%

    @app.callback(
        output=[
            Output(
                "content_main",
                "style",
            ),
            Output(
                "sidebar_left",
                "style",
            ),
            Output(
                "sidebar_right",
                "style",
            ),
            Output(
                "sidebar_left_state",
                "data",
            ),
            Output(
                "sidebar_right_state",
                "data",
            ),
            Output(
                "sidebar_left_toggle_btn",
                "outline",
            ),
            Output(
                "sidebar_right_toggle_btn",
                "outline",
            ),
        ],
        inputs=dict(
            _sidebar_left_toggle_btn=Input(
                "sidebar_left_toggle_btn",
                "n_clicks",
            ),
            _sidebar_right_toggle_btn=Input(
                "sidebar_right_toggle_btn",
                "n_clicks",
            ),
            current_state_sidebar_left=State(
                "sidebar_left_state",
                "data",
            ),
            current_state_sidebar_right=State(
                "sidebar_right_state",
                "data",
            ),
            sidebar_left_toggle_btn_outline=State(
                "sidebar_left_toggle_btn",
                "outline",
            ),
            sidebar_right_toggle_btn_outline=State(
                "sidebar_right_toggle_btn",
                "outline",
            ),
        ),
    )
    def toggle_sidebars(
        _sidebar_left_toggle_btn,
        _sidebar_right_toggle_btn,
        current_state_sidebar_left,
        current_state_sidebar_right,
        sidebar_left_toggle_btn_outline,
        sidebar_right_toggle_btn_outline,
    ):

        button_id = viz_utils.get_button_id(dash.callback_context)

        # if the toggle filter button was clicked
        if button_id == "sidebar_left_toggle_btn":
            return (
                *viz_utils.toggle_filter(
                    configurations,
                    current_state_sidebar_left,
                    current_state_sidebar_right,
                ),
                not sidebar_left_toggle_btn_outline,
                sidebar_right_toggle_btn_outline,
            )

        # if the toggle plot button was clicked
        elif button_id == "sidebar_right_toggle_btn":
            return (
                *viz_utils.toggle_plot(
                    configurations,
                    current_state_sidebar_left,
                    current_state_sidebar_right,
                ),
                sidebar_left_toggle_btn_outline,
                not sidebar_right_toggle_btn_outline,
            )

        # base configuration
        else:
            return *start_configuration, True, False

    #%%

    @app.callback(
        output=[
            Output("memory-store-csv", "data"),
            Output("memory-store-pdf", "data"),
        ],
        inputs=dict(
            navbar_btn_export_csv=Input(
                "navbar_btn_export_csv",
                "n_clicks",
            ),
            navbar_btn_export_pdf=Input(
                "navbar_btn_export_pdf",
                "n_clicks",
            ),
            sidebar_left_dropdown_samples=State(
                "sidebar_left_dropdown_samples",
                "value",
            ),
            sidebar_left_taxa_input_specific=State(
                "sidebar_left_taxa_input_specific",
                "value",
            ),
            sidebar_left_results_dynamic_value=State(
                {"type": "sidebar_left_results_dynamic", "index": ALL},
                "value",
            ),
            sidebar_left_results_dynamic_ids=State(
                {"type": "sidebar_left_results_dynamic", "index": ALL},
                "id",
            ),
            sidebar_left_taxa_input_path_contains=State(
                "sidebar_left_taxa_input_path_contains",
                "value",
            ),
            # prevent_initial_call=True,
        ),
    )
    def make_data_for_export(
        navbar_btn_export_csv,
        navbar_btn_export_pdf,
        sidebar_left_dropdown_samples,
        sidebar_left_taxa_input_specific,
        sidebar_left_results_dynamic_value,
        sidebar_left_results_dynamic_ids,
        sidebar_left_taxa_input_path_contains,
    ):

        ctx = dash.callback_context
        if not ctx.triggered:
            button_id = "No clicks yet"
        else:
            button_id = ctx.triggered[0]["prop_id"].split(".")[0]

        if navbar_btn_export_csv or navbar_btn_export_pdf:

            df_results_filtered = filter_dataframe(
                viz_results,
                sidebar_left_dropdown_samples,
                sidebar_left_results_dynamic_ids,
                sidebar_left_results_dynamic_value,
                sidebar_left_taxa_input_specific,
                sidebar_left_taxa_input_path_contains,
            )

            df_dict = df_results_filtered.to_dict("records")

            if button_id == "navbar_btn_export_csv":
                return df_dict, dash.no_update

            if button_id == "navbar_btn_export_pdf":
                return dash.no_update, df_dict

        else:
            raise PreventUpdate

    @app.callback(
        output=Output(
            "export_csv",
            "data",
        ),
        inputs=dict(
            data=Input(
                "memory-store-csv",
                "data",
            ),
        ),
    )
    def export_csv(
        data,
    ):

        # if data is None or n_clicks is None:
        if data is None:
            return dash.no_update

        df_results_filtered = pd.DataFrame(data)

        send_dataframe = dcc.send_data_frame(
            df_results_filtered.to_csv,
            "filtered_results.csv",
            index=False,
        )
        return send_dataframe

    @app.long_callback(
        output=Output(
            "export_pdf",
            "data",
        ),
        inputs=dict(
            data=Input(
                "memory-store-pdf",
                "data",
            ),
        ),
        running=[
            (
                Output("modal_pdf", "is_open"),
                True,
                False,
            ),
        ],
        progress=[
            Output(
                "progress_bar",
                "value",
            ),
            Output(
                "progress_bar",
                "max",
            ),
        ],
        cancel=[
            Input(
                "modal_pdf_close_button",
                "n_clicks",
            )
        ],
        manager=long_callback_manager,
        prevent_initial_call=True,
    )
    def export_pdf(
        set_progress,
        data,
    ):

        # if data is None or n_clicks is None:
        if data is None:
            return dash.no_update

        df_results_filtered = pd.DataFrame(data)

        pdf_path = "pdf_export.pdf"
        figures.save_pdf_plots(
            df_results_filtered,
            viz_results,
            pdf_path=pdf_path,
            set_progress=set_progress,
        )
        return dcc.send_file(pdf_path)

    # %%

    return app


#%%


def apply_sidebar_left_tax_id(viz_results, d_filter, sidebar_left_taxa_input_specific):

    d_filter = d_filter.copy()

    if (
        sidebar_left_taxa_input_specific is None
        or len(sidebar_left_taxa_input_specific) == 0
    ):
        return d_filter

    for tax in sidebar_left_taxa_input_specific:
        if tax in viz_results.all_tax_ids:
            d_filter = viz_utils.append_to_list_if_exists(d_filter, "tax_ids", tax)
        elif tax in viz_results.all_tax_names:
            d_filter = viz_utils.append_to_list_if_exists(d_filter, "tax_names", tax)
        elif tax in viz_results.all_tax_ranks:
            d_filter = viz_utils.append_to_list_if_exists(d_filter, "tax_ranks", tax)
        else:
            raise AssertionError(f"Tax {tax} could not be found. ")

    return d_filter


#%%


def filter_dataframe(
    viz_results,
    sidebar_left_dropdown_samples,
    sidebar_left_results_dynamic_ids,
    sidebar_left_results_dynamic_value,
    sidebar_left_taxa_input_specific,
    sidebar_left_taxa_input_path_contains,
):

    d_filter = {"samples": sidebar_left_dropdown_samples}

    columns_no_log = [id["index"] for id in sidebar_left_results_dynamic_ids]
    for sample, values in zip(columns_no_log, sidebar_left_results_dynamic_value):
        d_filter[sample] = values

    d_filter = apply_sidebar_left_tax_id(
        viz_results,
        d_filter,
        sidebar_left_taxa_input_specific,
    )

    df_results_filtered = viz_results.filter(
        d_filter,
        rank=sidebar_left_taxa_input_path_contains,
    )

    return df_results_filtered



#%%
import jax.numpy as jnp
import numpy as np
import numpyro
import pandas as pd
from jax import jit
from jax.random import PRNGKey as Key
from numba import njit
from numpyro import distributions as dist
from numpyro.infer import MCMC, NUTS, Predictive, log_likelihood
from scipy.special import logsumexp
from scipy.stats import beta as sp_beta
from scipy.stats import betabinom as sp_betabinom
from scipy.stats import norm as sp_norm

from metaDMG.fit import fit_utils


#%%

numpyro.enable_x64()

priors = fit_utils.get_priors()
A_prior = priors["A"]  # mean = 0.01, concentration = 1
q_prior = priors["q"]  # mean = 0.2, concentration = 5
c_prior = priors["c"]  # mean = 0.1, concentration = 10
phi_prior = priors["phi"]

#%%


def model_PMD(x, N, k=None):
    x_abs = jnp.abs(x)

    A = numpyro.sample("A", dist.Beta(A_prior[0], A_prior[1]))
    q = numpyro.sample("q", dist.Beta(q_prior[0], q_prior[1]))
    c = numpyro.sample("c", dist.Beta(c_prior[0], c_prior[1]))
    Dx = jnp.clip(numpyro.deterministic("Dx", A * (1 - q) ** (x_abs - 1) + c), 0, 1)

    delta = numpyro.sample("delta", dist.Exponential(1 / phi_prior[1]))
    phi = numpyro.deterministic("phi", delta + phi_prior[0])

    alpha = numpyro.deterministic("alpha", Dx * phi)
    beta = numpyro.deterministic("beta", (1 - Dx) * phi)

    numpyro.sample("obs", dist.BetaBinomial(alpha, beta, N), obs=k)


def model_null(x, N, k=None):
    c = numpyro.sample("c", dist.Beta(c_prior[0], c_prior[1]))
    Dx = numpyro.deterministic("Dx", c)
    delta = numpyro.sample("delta", dist.Exponential(1 / phi_prior[1]))
    phi = numpyro.deterministic("phi", delta + phi_prior[0])

    alpha = numpyro.deterministic("alpha", Dx * phi)
    beta = numpyro.deterministic("beta", (1 - Dx) * phi)

    numpyro.sample("obs", dist.BetaBinomial(alpha, beta, N), obs=k)


#%%


def filter_out_k(data):
    return {key: value for key, value in data.items() if key != "k"}


def is_model_PMD(model):
    name = model.__name__.lower()
    if "pmd" in name:
        return True
    elif "null" in name:
        return False
    raise AssertionError(f"Model should be PMD or null, got {model}")


#%%


@jit
def _get_posterior_PMD(rng_key, samples, *args, **kwargs):
    return Predictive(model_PMD, samples)(rng_key, *args, **kwargs)


@jit
def _get_posterior_null(rng_key, samples, *args, **kwargs):
    return Predictive(model_null, samples)(rng_key, *args, **kwargs)


def get_posterior_predictive(mcmc, data):
    posterior_samples = mcmc.get_samples()
    rng_key = Key(0)
    data_no_k = filter_out_k(data)
    if is_model_PMD(mcmc.sampler.model):
        return _get_posterior_PMD(rng_key, posterior_samples, **data_no_k)
    else:
        return _get_posterior_null(rng_key, posterior_samples, **data_no_k)


def get_posterior_predictive_obs(mcmc, data):
    return get_posterior_predictive(mcmc, data)["obs"]


#%%


def get_n_sigma_probability(n_sigma):
    return sp_norm.cdf(n_sigma) - sp_norm.cdf(-n_sigma)


CONF_1_SIGMA = get_n_sigma_probability(1)


# def compute_posterior(
#     mcmc, data, func_avg=np.mean, func_dispersion=lambda x: np.std(x, axis=0)
# ):
#     """func = central tendency function, e.g. np.mean or np.median"""
#     posterior_predictive = get_posterior_predictive(mcmc, data)
#     predictions_fraction = posterior_predictive["obs"] / data["N"]
#     y_average = func_avg(predictions_fraction, axis=0)
#     # y_dispersion = numpyro.diagnostics.hpdi(predictions_fraction, prob=0.68)
#     y_dispersion = func_dispersion(predictions_fraction)
#     return y_average, y_dispersion


def add_D_max_information(fit_result, mcmc, data):
    # posterior = get_posterior_predictive(mcmc, data)
    # c = mcmc.get_samples()["c"]
    # f = posterior["obs"] / data["N"]
    # f = f[:, 0]
    # D_max_samples = f - c
    # D_max_mu = np.mean(D_max_samples).item()
    # D_max_std = np.std(D_max_samples).item()

    # New method, more similar to frequentist and better when few reads
    A = mcmc.get_samples()["A"]
    # c = mcmc.get_samples()["c"]
    phi = mcmc.get_samples()["phi"]
    N = max(data["N"][0], 1)
    mu = np.mean(A)
    std = np.mean(np.sqrt(A * (1 - A) * (phi + N) / ((phi + 1) * N)))
    # mu.item(), std.item()

    Dx = A
    alpha = Dx * phi
    beta = (1 - Dx) * phi

    pdf = sp_betabinom(N, alpha, beta)
    # 1000x faster approximation for ppf
    pdf_approx = sp_betabinom(N, alpha.mean(), beta.mean())

    pdf_beta = sp_beta(alpha, beta)

    # pdf.pmf(0).mean()
    # pdf_approx.pmf(0) # 3 times faster for pmf

    prefix = "Bayesian_"
    fit_result[f"{prefix}D_max"] = mu.item()
    fit_result[f"{prefix}D_max_std"] = std.item()

    fit_result[f"{prefix}prob_lt_5p_damage"] = pdf_beta.cdf(0.05).mean()
    # fit_result[f"{prefix}prob_lt_5p_damage_betabinom"] = pdf.cdf(0.05 * N).mean()
    fit_result[f"{prefix}prob_lt_2p_damage"] = pdf_beta.cdf(0.02).mean()
    # fit_result[f"{prefix}prob_lt_2p_damage_betabinom"] = pdf.cdf(0.02 * N).mean()
    fit_result[f"{prefix}prob_lt_1p_damage"] = pdf_beta.cdf(0.01).mean()
    # fit_result[f"{prefix}prob_lt_1p_damage_betabinom"] = pdf.cdf(0.01 * N).mean()
    fit_result[f"{prefix}prob_lt_0.1p_damage"] = pdf_beta.cdf(0.001).mean()
    # fit_result[f"{prefix}prob_lt_0.1p_damage_betabinom"] = pdf.cdf(0.001 * N).mean()

    fit_result[f"{prefix}prob_zero_damage"] = pdf.cdf(0).mean()

    fit_result[f"{prefix}D_max_median"] = pdf.median().mean() / N

    for n_sigma in [1, 2, 3]:
        conf = get_n_sigma_probability(n_sigma)
        fit_result[f"{prefix}D_max_confidence_interval_{n_sigma}_sigma_low"] = (
            pdf_approx.ppf((1 - conf) / 2.0).mean() / N
        )
        fit_result[f"{prefix}D_max_confidence_interval_{n_sigma}_sigma_high"] = (
            pdf_approx.ppf((1 + conf) / 2.0).mean() / N
        )

    fit_result[f"{prefix}D_max_confidence_interval_95_low"] = (
        pdf_approx.ppf((1 - 0.95) / 2.0).mean() / N
    )
    fit_result[f"{prefix}D_max_confidence_interval_95_high"] = (
        pdf_approx.ppf((1 + 0.95) / 2.0).mean() / N
    )

    # fit_result["Bayesian_D_max"] = D_max["mu"]
    # fit_result["Bayesian_D_max_std"] = D_max["std"]
    # fit_result["Bayesian_D_max_median"] = D_max["median"]
    # fit_result["Bayesian_D_max_confidence_interval_1_sigma_low"] = D_max[
    #     "confidence_interval_1_sigma_low"
    # ]
    # fit_result["Bayesian_D_max_confidence_interval_1_sigma_high"] = D_max[
    #     "confidence_interval_1_sigma_high"
    # ]
    # fit_result["Bayesian_D_max_confidence_interval_95_low"] = D_max[
    #     "confidence_interval_95_low"
    # ]
    # fit_result["Bayesian_D_max_confidence_interval_95_high"] = D_max[
    #     "confidence_interval_95_high"
    # ]

    return fit_result


#%%


@jit
def _compute_log_likelihood_PMD(posterior_samples, data):
    return log_likelihood(model_PMD, posterior_samples, **data)["obs"]


@jit
def _compute_log_likelihood_null(posterior_samples, data):
    return log_likelihood(model_null, posterior_samples, **data)["obs"]


def compute_log_likelihood(mcmc, data):
    posterior_samples = mcmc.get_samples()
    if is_model_PMD(mcmc.sampler.model):
        return _compute_log_likelihood_PMD(posterior_samples, data)
    else:
        return _compute_log_likelihood_null(posterior_samples, data)


#%%


def get_lppd_and_waic(mcmc, data):
    d_results = {}
    # get the log likehood for each (point, num_samples)
    logprob = np.asarray(compute_log_likelihood(mcmc, data))
    # lppd for each observation
    lppd_i = logsumexp(logprob, 0) - np.log(logprob.shape[0])
    d_results["lppd_i"] = lppd_i
    # lppd
    lppd = lppd_i.sum()
    d_results["lppd"] = lppd
    # waic penalty for each observation
    pWAIC_i = np.var(logprob, 0)
    d_results["pWAIC_i"] = pWAIC_i
    # waic penalty # the effective number of parameters penalty
    pWAIC = pWAIC_i.sum()
    d_results["pWAIC"] = pWAIC
    # waic  for each observation
    waic_i = -2 * (lppd_i - pWAIC_i)
    d_results["waic_i"] = waic_i
    # waic # prediction of  out-of-sample deviance
    waic = waic_i.sum()
    d_results["waic"] = waic
    # standard error of waic
    # waic_vec = -2 * (lppd_i - pWAIC_i)
    # waic_uncertainty = jnp.sqrt(logprob.shape[1] * jnp.var(waic_vec))
    return d_results


#%%


# def get_mean_of_variable(mcmc, variable, axis=0):
#     values = mcmc.get_samples()[variable]
#     return np.mean(values, axis=axis).item()


# def get_std_of_variable(mcmc, variable, axis=0):
#     values = mcmc.get_samples()[variable]
#     return np.std(values, axis=axis).item()


def add_summary_of_variable(
    fit_result,
    mcmc,
    variable,
    prefix="Bayesian_",
    axis=0,
):
    values = np.array(mcmc.get_samples()[variable])

    s = f"{prefix}{variable}"
    q_low = (1 - CONF_1_SIGMA) / 2.0
    q_high = (1 + CONF_1_SIGMA) / 2.0

    fit_result[f"{s}"] = np.mean(values, axis=axis)
    fit_result[f"{s}_std"] = np.std(values, axis=axis)
    fit_result[f"{s}_median"] = np.median(values, axis=axis)
    fit_result[f"{s}_confidence_interval_1_sigma_low"] = np.quantile(
        values, q_low, axis=axis
    )
    fit_result[f"{s}_confidence_interval_1_sigma_high"] = np.quantile(
        values, q_high, axis=axis
    )


def compute_rho_Ac(mcmc):
    A = mcmc.get_samples()["A"]
    c = mcmc.get_samples()["c"]
    return np.corrcoef(A, c)[0, 1]


@njit
def compute_dse(waic_i_x, waic_i_y):
    N = len(waic_i_x)
    return np.sqrt(N * np.var(waic_i_x - waic_i_y))


@njit
def compute_z_from_waic_is(waic_i_x, waic_i_y):
    dse = compute_dse(waic_i_x, waic_i_y)
    d_waic = waic_i_y.sum() - waic_i_x.sum()
    z = d_waic / dse
    return z


def compute_z(d_results_PMD, d_results_null):
    z = compute_z_from_waic_is(d_results_PMD["waic_i"], d_results_null["waic_i"])
    return z


@njit
def compute_z_jackknife_error_from_waic_is(waic_i_x, waic_i_y):
    N = len(waic_i_x)
    all_ids = np.arange(N)
    zs = np.empty(N)
    for i in range(N):
        mask = all_ids != i
        zs[i] = compute_z_from_waic_is(waic_i_x[mask], waic_i_y[mask])
    z_jack_std = np.std(zs)
    return z_jack_std


def compute_z_jackknife_error(d_results_PMD, d_results_null):
    return compute_z_jackknife_error_from_waic_is(
        d_results_PMD["waic_i"],
        d_results_null["waic_i"],
    )


#%%


def init_mcmc(model, **kwargs):

    mcmc_kwargs = dict(
        progress_bar=False,
        num_warmup=500,
        num_samples=1000,
        num_chains=1,  # problem when setting to 2
        chain_method="sequential",
        # http://num.pyro.ai/en/stable/_modules/numpyro/infer/mcmc.html#MCMC
    )

    return MCMC(NUTS(model), jit_model_args=True, **mcmc_kwargs, **kwargs)


def init_mcmc_PMD(**kwargs):
    mcmc_PMD = init_mcmc(model_PMD, **kwargs)
    return mcmc_PMD


def init_mcmc_null(**kwargs):
    mcmc_null = init_mcmc(model_null, **kwargs)
    return mcmc_null


def init_mcmcs(config, **kwargs):
    if config["bayesian"]:
        mcmc_PMD = init_mcmc_PMD(**kwargs)
        mcmc_null = init_mcmc_null(**kwargs)
    else:
        mcmc_PMD = None
        mcmc_null = None
    return mcmc_PMD, mcmc_null


def fit_mcmc(mcmc, data, seed=0):
    mcmc.run(Key(seed), **data)


def use_last_state_as_warmup_state(mcmc):
    # https://github.com/pyro-ppl/numpyro/issues/539
    mcmc._warmup_state = mcmc._last_state


def add_Bayesian_fit_result(
    fit_result,
    data,
    mcmc_PMD,
    mcmc_null,
):

    d_results_PMD = get_lppd_and_waic(mcmc_PMD, data)
    d_results_null = get_lppd_and_waic(mcmc_null, data)

    add_D_max_information(fit_result, mcmc_PMD, data)

    add_summary_of_variable(fit_result, mcmc_PMD, "A")
    add_summary_of_variable(fit_result, mcmc_PMD, "q")
    add_summary_of_variable(fit_result, mcmc_PMD, "c")
    add_summary_of_variable(fit_result, mcmc_PMD, "phi")

    fit_result["Bayesian_rho_Ac"] = compute_rho_Ac(mcmc_PMD)

    z = compute_z(d_results_PMD, d_results_null)
    fit_result["Bayesian_z"] = z


def make_fits(fit_result, data, mcmc_PMD, mcmc_null):
    fit_mcmc(mcmc_PMD, data)
    fit_mcmc(mcmc_null, data)  # do not use non-PMD model later on
    add_Bayesian_fit_result(fit_result, data, mcmc_PMD, mcmc_null)
    # mcmc_PMD.print_summary(prob=0.68)
    # mcmc_null.print_summary(prob=0.68)

    # if False:
    #     use_last_state_as_warmup_state(mcmc_PMD)
    #     use_last_state_as_warmup_state(mcmc_null)
